"""
User-defined JSONB filters.
"""

from typing import Any

from sqlalchemy import Float, Integer, String, and_, func, or_
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.operators import custom_op, eq, ge, gt, in_op, le, like_op, lt, ne, notin_op, notlike_op  # noqa
from sqlalchemy.sql.type_api import TypeEngine
from vlutils.structures.dataclasses import dataclass

exists = custom_op("?")

META_OPERATOR_MAP = {
    "eq": eq,
    "neq": ne,
    "like": like_op,
    "nlike": notlike_op,
    "in": in_op,
    "nin": notin_op,
    "gt": gt,
    "gte": ge,
    "lt": lt,
    "lte": le,
}

META_TYPE_MAP = {
    "string": String,
    "integer": Integer,
    "numeric": Float,
}


@dataclass(withSlots=True)
class UDFilter:
    """User-defined JSONB filter."""

    path: list[str]
    dtype: TypeEngine
    operator: callable
    value: Any

    def jsonb_extract(self, data, nullable: bool = False) -> ColumnElement:
        """Make value extraction."""
        if self.dtype is String or nullable:
            return func.jsonb_extract_path_text(data, *self.path)
        return func.jsonb_extract_path_text(data, *self.path).cast(self.dtype)

    def jsonb_exists(self, data) -> ColumnElement:
        """Make existence rule."""
        data = func.jsonb_extract_path(data, *dpath) if (dpath := self.path[:-1]) else data
        return data.operate(exists, self.path[-1])

    def operate(self, data) -> ColumnElement:
        """Make comparison operation."""
        if self.operator is eq and self.value is None:
            return and_(
                self.jsonb_exists(data),
                self.jsonb_extract(data).operate(self.operator, self.value),
            )
        if self.operator is in_op and None in self.value:
            return or_(
                and_(self.jsonb_exists(data), self.jsonb_extract(data, nullable=True).is_(None)),
                self.jsonb_extract(data).operate(self.operator, filter(None, self.value)),
            )
        if self.operator is notin_op and None in self.value:
            return and_(
                self.jsonb_extract(data, nullable=True).isnot(None),
                self.jsonb_extract(data).operate(self.operator, filter(None, self.value)),
            )
        return self.jsonb_extract(data).operate(self.operator, self.value)  # noqa
